import uuid
from azure.data.tables import TableServiceClient


class NoConnectionToDB(Exception):
    pass


class TableHandler():
    def __init__(self, connection_str):
        try:
            table_service = TableServiceClient.from_connection_string(conn_str=connection_str)
        except Exception:
            raise NoConnectionToDB("Table Service Client could not be reached.")
        self._table_client = table_service.get_table_client(table_name="UserRates")

    def query_user_id(self, user_id: str):
        return self._query_entities_in_handler_data("UserID", user_id)

    def query_offer_id(self, offer_id: str):
        return self._query_entities_in_handler_data("OfferID", offer_id)

    def increment_num_calls_handler(self, offer_id: str):
        entities = self.query_offer_id(offer_id)
        self._increment_calls("NumCallsHandler", entities[0])

    def increment_num_calls_buddy_generator(self, offer_id: str):
        entities = self.query_offer_id(offer_id)
        self._increment_calls("NumCallsBuddyGenerator", entities[0])

    def create_entity(self, offer_id: str, user_id: str):
        entity = {
            "PartitionKey": "HandlerData",
            "RowKey": str(uuid.uuid4()),
            "NumCallsHandler": 0,
            "OfferID": offer_id,
            "UserID": user_id
        }
        return self._table_client.create_entity(entity)

    def _increment_calls(self, key: str, entity: dict):
        if entity:
            entity[key] += 1
            self._table_client.update_entity(entity)

    def _query_entities_in_handler_data(self, key: str, value: str):
        filters = f"PartitionKey eq 'HandlerData' and {key} eq '{value}'"
        return list(self._table_client.query_entities(filters))



import json
import threading

import paho.mqtt.client as mqtt


class MQTTClient:
    def __init__(self, client_id, logger, endpoint, port=None):
        def on_connect(client, userdata, connect_flags, reason_code, properties):
            self.logger.info('CONNACK received', reason_code=reason_code, properties=properties)
            self._connected.set()

        self.logger = logger
        self._endpoint = endpoint
        self._port = int(port)
        self._connected = threading.Event()
        self._session_id = None
        self._request_id = None
        self._client_id = client_id

        self._client = mqtt.Client(
            mqtt.CallbackAPIVersion.VERSION2,
            transport="websockets",
            reconnect_on_failure=True,
            clean_session=True,
            client_id=client_id
        )
        self._client.on_connect = on_connect
        self._client.tls_set()
        self._message_counter = 0
        self._streamed = []

    def start(self):
        self._client.connect(self._endpoint, self._port)
        self._client.loop_start()

    @property
    def session_id(self):
        return self._session_id

    @session_id.setter
    def session_id(self, session_id):
        self._session_id = session_id

    @property
    def request_id(self):
        return self._request_id

    @request_id.setter
    def request_id(self, request_id):
        self._request_id = request_id

    @property
    def streamed(self):
        return self._streamed

    @property
    def topics(self):
        yield f'tm/id/{self.session_id}'
        if self.request_id:
            yield f'tm/id/{self.session_id}-{self.request_id}'

    def stream_utterance(self, persona=None, voice=None, utterance="", status=None):
        self.set_persona(persona)
        self.set_voice(voice)
        self.stream_chunk(utterance + " ")

    def set_persona(self, persona):
        self._stream_to_frontend({"event": "STREAMING_SET_PERSONA", "data": persona if persona else ""})

    def set_voice(self, voice):
        self._stream_to_frontend({"event": "STREAMING_SET_VOICE", "data": voice if voice else ""})

    def stream_chunk(self, chunk):
        self._stream_to_frontend({"event": "STREAMING_CHUNK", "data": chunk})
        self._streamed.append(chunk)

    def prepare_stream(self):
        self._message_counter = 0
        self._streamed = []

    def end_stream(self):
        self._stream_to_frontend({"event": "STREAMING_DONE"})

    def _stream_to_frontend(self, message):
        self._message_counter += 1
        message |= {"id": f"{self._message_counter}_{self._client_id}"}
        self.logger.debug("streaming to frontend", message=message, session_id=self.session_id)
        self._connected.wait()
        for topic in self.topics:
            self._client.publish(topic, json.dumps(message))
